﻿Imports System.ServiceModel
Imports System.ServiceModel.Channels
Imports System.ServiceModel.Description

Module Module1

    Private _binding As New NetNamedPipeBinding

    Sub Main()

        ' Connect1: 特定のエンドポイントアドレスで接続を待つ。
        Dim listener1 As IChannelListener(Of IDuplexSessionChannel) _
            = _binding.BuildChannelListener(Of IDuplexSessionChannel)( _
                New Uri("net.pipe://localhost/IDuplexSessionChannel1/ConnectionInfo"), _
                New BindingParameterCollection() _
            )

        Try
            listener1.Open()
        Catch ex As AddressAlreadyInUseException
            Throw
        Catch ex As TimeoutException
            Throw
        Catch ex As CommunicationException
            Throw
        End Try

        Do

            Dim connectionChannel As IDuplexSessionChannel = CreateOpenedChannel(listener1)    ' block

            Dim listener2 As IChannelListener(Of IDuplexSessionChannel)
            With Nothing
                ' Connect2-1: 動的なエンドポイントを作成
                listener2 = _binding.BuildChannelListener(Of IDuplexSessionChannel)( _
                    New Uri("net.pipe://localhost"), _
                    "IDuplexSessionChannel1", _
                    ListenUriMode.Unique, _
                    New BindingParameterCollection() _
                )

                ' エンドポイントアドレスが net.pipe://COMPUTER_NAME/GUID になる例
                'Dim listener2 As IChannelListener(Of IDuplexSessionChannel) _
                '    = _binding.BuildChannelListener(Of IDuplexSessionChannel)( _
                '        New EndpointAddress("net.pipe://localhost"), _
                '        New BindingParameterCollection() _
                '    )

                Try
                    listener2.Open()
                Catch ex As AddressAlreadyInUseException
                    Throw
                Catch ex As TimeoutException
                    Throw
                Catch ex As CommunicationException
                    Throw
                End Try

                ' 動的なエンドポイントアドレスをクライアントに送信
                Dim newEndpointUri As String = listener2.Uri.AbsoluteUri
                Dim reply As Message = Message.CreateMessage(_binding.MessageVersion, "net.pipe://localhost/IDuplexSessionChannel1/ConnectionInfo", newEndpointUri)
                __Out("connectionChannel.Send()")
                connectionChannel.Send(reply)
                __Out("connectionChannel.Sent: " & newEndpointUri)
                '__Out("connectionChannel.Sent: " & reply.GetBody(Of String)())   ' System.InvalidOperationException: "このメッセージは書き込まれているため、操作をサポートできません。"
                'reply.Close()
            End With

            ' データ転送用スレッド開始
            Dim t As New Threading.Thread(AddressOf Module1.TransportThread)
            t.Start(listener2)

            Try
                connectionChannel.Close()
            Catch ex As TimeoutException
                Throw
            Catch ex As CommunicationException
                Throw
            End Try

        Loop

        listener1.Close()

    End Sub

    Sub TransportThread(listener2 As IChannelListener(Of IDuplexSessionChannel))
        __Out("Start Transport: " & listener2.Uri.AbsoluteUri)

        ' Connect2-2: 動的なエンドポイントで接続を待つ。
        Dim transportChannel As IDuplexSessionChannel = CreateOpenedChannel(listener2)    ' block

        With Nothing
            ' Transport2-1
            Dim str As String = ">"
            Dim msg As Message = Message.CreateMessage(_binding.MessageVersion, listener2.Uri.AbsoluteUri, str)
            __Out("transportChannel.Send1()")
            transportChannel.Send(msg)
            __Out("transportChannel.Sent1: " & str)
            'reply.Close()
        End With

        With Nothing
            ' Transport2-2
            __Out("transportChannel.Receive1()")
            Dim received As Message = transportChannel.Receive()
            Dim receivedString As String = received.GetBody(Of String)()
            __Out("transportChannel.Received1: " & receivedString)
            received.Close()

            Dim str As String = "Re:" & receivedString
            Dim msg As Message = Message.CreateMessage(_binding.MessageVersion, listener2.Uri.AbsoluteUri, str)
            __Out("transportChannel.Send2()")
            transportChannel.Send(msg)
            __Out("transportChannel.Sent2: " & str)
            'reply.Close()
        End With

        ' 動的なエンドポイントでの通信終了
        Try
            transportChannel.Close()
        Catch ex As TimeoutException
            Throw
        Catch ex As CommunicationException
            Throw
        End Try

        Try
            listener2.Close()
        Catch ex As TimeoutException
            Throw
        Catch ex As CommunicationException
            Throw
        End Try

        __Out("End Transport: " & listener2.Uri.AbsoluteUri)
        __Out("")

    End Sub

    Function CreateOpenedChannel(listener As IChannelListener(Of IDuplexSessionChannel)) As IDuplexSessionChannel
        Dim channel As IDuplexSessionChannel = Nothing

        Do Until Not IsNothing(channel) AndAlso channel.State = CommunicationState.Opened
            Try
                channel = listener.AcceptChannel()  ' クライアントが接続するまで停止
                channel.Open()
            Catch ex As TimeoutException
                __Out(ex.Message)
            Catch ex As CommunicationException
                Throw
            End Try
        Loop

        Debug.Assert(Not IsNothing(channel))
        Return channel
    End Function

    Public Sub __Out(message As String)
        Debug.AutoFlush = True
        Debug.WriteLine(My.Application.Info.AssemblyName & ": " & Now.ToString() & " " & message)
        Console.WriteLine(message)
    End Sub

End Module
